@@ -1,11 +1,3 @@
-//  NormNet: a Deep Artificial Neural Network using normalizing neurons to investigate alternatives to bias
-//  based expressivity. Gradient based connection-wise dropout is also included due to the numeric properties
-//  inherent to normalization (thanks to a little trig and calculus). This allows NormNet to also be a sparse-searching
-//  "top down" architecture search where the parameters are recoverable thanks again to the gradient.
-//  It is hypothesized and investigated here that the aforementioned features will lead to much more robust transfer
-//  learning and therein generalization. The only significant downsides to this is a division by constant operation on each node
-//  (seemingly/empirically not as bad as it sounds).
-
 //TODO: allow this to fallback to TF-CPU for machines without GPU and possibly just do inference with a build script.
 //TODO: floats stink norm net is an attempt to realizing rot net and uint8 operations with bounded functions while retaining all the flaws in DNNs.. (go back to rot net at some point)
 //TODO: rayon all the things
@@ -13,12 +5,15 @@
 
 //TODO: type errors for anything other than u64 architecture due to casts. 
 //      Also precision errors for summing f32s which should accumulate to f64.
-//TODO: logging for effeciency, SED println! to a logging crate since stdout logs are already formatted well
+//TODO: logging for effeciency, SED log::debug! to a logging crate since stdout logs are already formatted well
 //NOTE: DONE WITH FEATURES. this is a basic layers based tensorflow backed ANN architecture framework. 
 //      Anything else should be written outside this crate and integrated only if general enough.
 
 //allow unstable features
 // #![feature(int_log)]
+mod layers;
+
+use log;
 use serde::{Serialize, Deserialize};
 use serde_derive::{Serialize, Deserialize};
 use serde_json::{Value};
@@ -71,195 +66,10 @@ use tensorflow::REGRESS_INPUTS;
 use tensorflow::REGRESS_METHOD_NAME;
 use tensorflow::REGRESS_OUTPUTS;
 use uuid::Uuid;
+use layers::*;
 
-/// A standard fully connected layer with bias term
-///
-/// `activation` is a function which takes a tensor and applies an activation
-/// function such as sigmoid.
-///
-/// Returns variables created and the layer output.
-fn layer<O1: Into<Output>>(
-    input: O1,
-    input_size: u64,
-    output_size: u64,
-    activation: &dyn Fn(Output, &mut Scope) -> Result<Output, Status>,
-    scope: &mut Scope,
-) -> Result<(Vec<Variable>, Output), Status> {
-    let mut scope = scope.new_sub_scope("layer");
-    let scope = &mut scope;
-    let w_shape = ops::constant(&[input_size as i64, output_size as i64][..], scope)?;
-    let w = Variable::builder()
-        .initial_value(
-            ops::RandomStandardNormal::new()
-                .dtype(DataType::Half)
-                .build(w_shape, scope)?,
-        )
-        .data_type(DataType::Half)
-        .shape([input_size, output_size])
-        .build(&mut scope.with_op_name("w"))?;
-    let b = Variable::builder()
-        .const_initial_value(Tensor::<f16>::new(&[output_size]))
-        .build(&mut scope.with_op_name("b"))?;
-    //n is input_size to be divided at each node in order to normalize the signals at each node before activation
-    Ok((
-        vec![w.clone(), b.clone()],
-        activation(
-            ops::add(
-                ops::mat_mul(input, w.output().clone(), scope)?,
-                b.output().clone(),
-                scope,
-            )?
-            .into(),
-            scope,
-        )?,
-    ))
-}
-
-pub fn norm_layer<O1: Into<Output>>(
-    input: O1,
-    input_size: u64,
-    output_size: u64,
-    activation: &dyn Fn(Output, &mut Scope) -> Result<Operation, Status>,
-    scope: &mut Scope,
-) -> Result<(Vec<Variable>, Output, Operation), Status> {
-    let mut scope = scope.new_sub_scope("layer");
-    let scope = &mut scope;
-    let w_shape = ops::constant(&[input_size as i64, output_size as i64][..], scope)?;
-    let w = Variable::builder()
-        .initial_value(
-            ops::RandomStandardNormal::new()
-                .dtype(DataType::Float)
-                .build(w_shape.clone(), scope)?,
-        )
-        .data_type(DataType::Float)
-        .shape([input_size, output_size])
-        .build(&mut scope.with_op_name("w"))?;
-
-    let n = ops::constant(input_size as f32, scope)?;
-
-    //NOTE: tan on weights is to force weights to dropout but use the gradient for better dropout than random node based dropout
-    //NOTE: we multiply the activation by 100 to represent values </>than 1/-1
-    let output_op = ops::div(
-        ops::mat_mul(
-            input,
-            //this sets the gradients to dropout weights
-            ops::tan(w.output().clone(), scope)?,
-            scope,
-        )?,
-        n,
-        scope,
-    )?;
-
-    let act = activation(
-        //this normalizes to speed up trainning and sample efficiency
-        output_op.into(),
-        scope,
-    )?;
-    let output_op = act.clone();
-    // .into(); //,
-    Ok((vec![w.clone()], act.into(), output_op))
-}
-
-///a normal layer as above but with residual connections
-fn norm_layer_res<O1: Into<Output>>(
-    input: O1,
-    res_input: O1,
-    input_size: u64,
-    output_size: u64,
-    activation: &dyn Fn(Output, &mut Scope) -> Result<Output, Status>,
-    scope: &mut Scope,
-) -> Result<(Vec<Variable>, Output), Status> {
-    let mut scope = scope.new_sub_scope("layer");
-    let scope = &mut scope;
-    let w_shape = ops::constant(&[input_size as i64, output_size as i64][..], scope)?;
-    let w = Variable::builder()
-        .initial_value(
-            ops::RandomStandardNormal::new()
-                .dtype(DataType::Half)
-                .build(w_shape.clone(), scope)?,
-        )
-        .data_type(DataType::Half)
-        .shape([input_size, output_size])
-        .build(&mut scope.with_op_name("w"))?;
-
-    //n is input_size to be divided at each node in order to normalize the signals at each node before activation
-    let input_size = 2 * input_size;
-    //TODO: concat the input and res tensors
-    // let concat = ops::concat(0,vec![input, res_input],  scope)?.into();
-
-    let scalar_coe = ops::constant(f16::from_f32(0.1), scope)?;
-    let cur = ops::mat_mul(
-        input,
-        //NOTE: division for half stability the higher this value the more stable the trainning but the less expressivity (domain) of the weights
-        //NOTE: tan on weights is to force weights to dropout but use the gradient for better dropout than random node based dropout
-        ops::multiply(
-            ops::tan(w.output().clone(), scope)?,
-            scalar_coe.clone(),
-            scope,
-        )?,
-        scope,
-    )?;
-
-    // let cur_res = ops::mat_mul(
-    // res_input,
-    // ops::multiply(ops::tan(w_res.output().clone(), scope)?, scalar_coe.clone(), scope)?,
-    // scope,
-    // )?;
-    // ops::tan(w_res.output().clone(), scope)?, scope)?;
-    let res_input = ops::multiply(scalar_coe, res_input, scope)?;
-    let cur = ops::add(cur, res_input, scope)?;
-
-    let n = ops::constant(f16::from_f32(input_size as f32), scope)?;
-
-    let res = activation(ops::div(cur, n, scope)?.into(), scope)?.into(); //,
-
-    // Ok((vec![w.clone(), w_res.clone()], res))
-    Ok((vec![w.clone()], res))
-}
 //TODO: defaults such as learning rate
-//================
-//----NORM_NET----
-//================
-/// A standard fully connected layer without bias trainnable parameters
-/// instead normalizing and dropping out connections at each node.
-///
-/// NOTE: currently inputs and outputs must be flattened if representing >1 dim data
-///
-/// #
-/// PROS:
-///
-/// * better exploration by removing instabilities inherent to bias
-///        
-/// * gradient based connection-wise dropout with tan weights
-///       
-/// * better transfer learning by removing bias connections
-///       
-/// * ~shouldn't~ have exploding gradient although vanishing gradient may be possible due to normalizing division.
-///      
-/// * technically connection wise dropout is divisive (top down) architecture search (e.g.: the opposite of NEAT (bottom up) which is agglomerative)
-///
-/// CONS:
-/// * may be slower due to more operations (division)
-///       
-/// * input and output must/should be tailored for normalized input/output (standard data science practices)
-///       
-/// * needs large type precision for stability, but the stability can be tuned (as apposed to bias which needs architectural considerations)
-///
-/// #
-/// NOTE: parameters goes to zero whereas biases find some 1-dimensional partition from -inf to inf. This helps
-/// build subgraph search modules (subtrees essentially). That can quickly optimize for distinct domains and labels via dropout.
-///
-/// NOTE:
-/// Tanh should the first and last layers activation function to map inputs and outputs to negative values.
-/// multiplying the output by some multiple of 10 allows the otherwise normalized network to take in and output whole integers.
-/// -x > y > x | x > 1
-///
-/// NOTE:
-/// We dont use BFloat since the integer range is only used as a buffer for addition overflow in matmul.
-/// In all other operations we are strictly bounded -1 > x > 1. As long as layer_width is not
-/// greater than Float range we are fine in the worst case (summing all 1's).
-/// Otherwise decimal precision of float type is our parameter type precision.
-pub struct NormNet<'a> {
+pub struct Brain<'a> {
     //TODO: refactor with builder pattern for RL, checkpointing and feedback network once tested and working.
     /// Tensorflow objects for user abstraction from Tensorflow
     scope: Scope,
@@ -301,44 +111,25 @@ pub struct NormNet<'a> {
     ///the current name of the checkpoint being searched
     checkpoint_name: Option<String>,
 }
-///The serialized representation of the model outside of tensorflow graph.
-///Primarily used for checkpoint search meta-learner.
-#[derive(Serialize, Deserialize,Debug)]
-struct SerializedNetwork{
-    /// The previous checkpoint that created this checkpoint for informed graph 
-    /// selection checkpoint search.
-    parent_search_name: String,
-    checkpoint_name: Option<String>,
-    lowest_error: f32,
-} 
-impl SerializedNetwork{
-    /// Constructs a new serialized network from a checkpoint name and the lowest error
-    fn new(parent_search_name: String, checkpoint_name: Option<String>, lowest_error: f32) -> Self{
-        Self{
-            parent_search_name,
-            checkpoint_name,
-            lowest_error,
-        }
-    }
-    fn restore(self, norm_net: &mut NormNet) {
-        // let mut NormNet = NormNet.clone();
-        norm_net.checkpoint_name = self.checkpoint_name;
-        norm_net.lowest_error = self.lowest_error;
-        // NormNet
-    }
-}
-impl <'a>NormNet <'a>{
+impl <'a>Brain <'a>{
     //TODO: type safety: some of the meta-learning values should be 64 bit for overflow since we sum 32s etc.
-    pub fn new(
+    pub fn new<Architecture>(
         name: &'a str,
         input_size: u64,
         output_size: u64,
+        //Number of neurons in each layer
         layer_width: u64,
+        //Number of layers in the network
         layer_height: u64,
+        layer: Architecture,
         max_integer: u32,
         learning_rate: f32,
         error_power: f32,
-    ) -> Result<NormNet<'a>, Status> {
+    ) -> Result<Brain<'a>, Status> 
+    where
+        Architecture: 
+            Fn(Output, u64, u64, &dyn Fn(Output, &mut Scope) -> Result<Operation, Status>,&mut Scope) -> 
+            Result<(Vec<Variable>,Output, Operation), Status>{
         assert!(max_integer % 10 == 0 || max_integer == 1, "max_integer must be a multiple of 10 or 1 since it represents order of magnitude of the integer range");
         let mut scope = Scope::new_root_scope();
 
@@ -353,45 +144,37 @@ impl <'a>NormNet <'a>{
             .build(&mut scope.with_op_name("label"))?;
 
         //CONSTRUCT NETWORK
-        //TODO: EXTRACT THIS TO LAYERS.RS. THIS IS ALL THATS NEEDED TO MAKE THIS AN ARCHITECTURE DESIGN CRATE.
+        //TODO: extract this to a builder pattern for constructor readability.
         let mut net_vars = vec![];
         let mut net_layers = vec![];
 
         //initial layer
-        let (vars, layer, _) = norm_layer(
-            Input.clone(),
-            input_size,
-            layer_width,
-            &|x, scope| Ok(ops::tanh(x, scope)?.into()),
-            &mut scope,
-        )?;
-        net_vars.extend(vars);
-        net_layers.push(layer.clone());
+        let (vars, cur_layer, _) = 
+        //TODO: also make activation functions an enum mod 
+        layer(Input.clone().into(), input_size, layer_width, &|x, scope| Ok(ops::tanh(x,scope)?.into()), &mut scope)?;
 
-        let mut prev_layer = layer;
+        net_vars.extend(vars);
+        net_layers.push(cur_layer.clone());
+
+        let mut prev_layer = cur_layer;
         //hidden layers
         for i in 0..layer_height - 2 {
-            let (vars, layer, _) = norm_layer(
-                prev_layer.clone(),
-                layer_width,
-                layer_width,
-                //NOTE: originally designed with tan but vanishing gradient can occur
-                &|x, scope| Ok(ops::tanh(x, scope)?.into()),
-                &mut scope,
-            )?;
-            prev_layer = layer.clone();
+            let (vars, cur_layer, _) = layer(prev_layer.clone().into(), layer_width, layer_width, &|x, scope| Ok(ops::tanh(x,scope)?.into()),&mut scope)?;
+            prev_layer = cur_layer.clone();
 
             net_vars.extend(vars);
-            net_layers.push(layer.clone());
+            net_layers.push(cur_layer.clone());
         }
 
-        //the final output layer is tanh to express negative values and multiplied to stabilize the
-        //half precision gradient as well as express whole integers outside of -1 and 1.
-        let (vars, output, Output_op) = norm_layer(
+        let (vars, output, Output_op) = 
+        layer(
             net_layers.last().unwrap().clone(),
             layer_width,
             output_size,
             &|x, scope| {
+                //TODO: extract to activation mod
+                //the final output layer is tanh to express negative values and multiplied to stabilize the
+                //half precision gradient as well as express whole integers outside of -1 and 1.
                 Ok(ops::multiply(
                     ops::tanh(x, scope)?,
                     ops::constant(max_integer as f32, scope)?,
@@ -401,6 +184,7 @@ impl <'a>NormNet <'a>{
             },
             &mut scope,
         )?;
+        //NOTE: need net_vars and net_layers as return values
         net_vars.extend(vars);
         net_layers.push(output.clone());
         //END OF CONSTRUCTING NETWORK
@@ -461,7 +245,7 @@ impl <'a>NormNet <'a>{
         }
         session.run(&mut run_args)?;
 
-        let mut init_norm_net = NormNet {
+        let mut init_norm_net = Brain {
             name,
             scope,
             session: session,
@@ -481,7 +265,7 @@ impl <'a>NormNet <'a>{
             checkpoint_name: None,
         };
 
-        //initialize NormNet save directory
+        //initialize Brain save directory
         fs::create_dir_all(init_norm_net.name).unwrap();
         init_norm_net.save().unwrap();
 
@@ -504,14 +288,14 @@ impl <'a>NormNet <'a>{
         // create a serialized_network object
         let serialized_network = SerializedNetwork{
             parent_search_name: name.clone(),
-            //TODO: see if we can remove this option in NormNet
+            //TODO: see if we can remove this option in Brain
             checkpoint_name: Some(dir.clone()),
             lowest_error: self.lowest_error,
         };
 
         //save the parent_search_name to dir which is where the model is saved, this is the edge to the checkpoint tree
         let file_name = format!("{}/checkpoint_data.json", dir);
-        println!("serializing non-tensorflow graph variables: {}", file_name);
+        log::debug!("serializing non-tensorflow graph variables: {}", file_name);
         // append the file type .j
         // create the file
         let mut file = fs::File::create(file_name.clone())?;
@@ -530,11 +314,11 @@ impl <'a>NormNet <'a>{
         assert!(checkpoint_window_size > 0, "checkpoint_window_size must be > 0");
         // let mut result = false;
 
-        // println!("lowest error: {}", self.lowest_error);
+        // log::debug!("lowest error: {}", self.lowest_error);
         // print window length and checkpoint_window_size
 
-        // println!("window length: {}", self.checkpoint_window.len());
-        // println!("checkpoint_window_size: {}", checkpoint_window_size);
+        // log::debug!("window length: {}", self.checkpoint_window.len());
+        // log::debug!("checkpoint_window_size: {}", checkpoint_window_size);
         //TODO: should we use checkpoint window size? there is a possibility of destroying the 
         //      best error in the window and still representing the now diverged solution.
         // the latent aspect of the gradient makes the solution still viable. it is "near" the 
@@ -544,23 +328,23 @@ impl <'a>NormNet <'a>{
         // checkpoint_window, as the name implies, can be modeled as a smoothing/clustering function.
 
         if self.checkpoint_window.len() >= checkpoint_window_size as usize {
-            // println!("evaluation window is full");
+            // log::debug!("evaluation window is full");
             // average checkpoint_window (dont need to divide since always comparing this value relatively) but we do so for more intuitive metrics relative to per column error
             let avg = self.checkpoint_window.iter().fold(0.0, |acc, x| acc + x)/ self.checkpoint_window.len() as f32;
             // print!("avg: {}", avg);
-            // println!(" vs lowest error: {}", self.lowest_error);
+            // log::debug!(" vs lowest error: {}", self.lowest_error);
 
             // this will probably get LICM'd anyways:
             let error_sum = error.iter().fold(0.0, |acc, x| acc + x)/error.len() as f32;
             self.checkpoint_window.push(error_sum);
 
-            // println!("error sum: {}", error_sum);
+            // log::debug!("error sum: {}", error_sum);
             if self.lowest_error > avg {
-                println!("NEW LOWEST ERROR: {}", avg);
+                log::debug!("NEW LOWEST ERROR: {}", avg);
                 self.lowest_error = avg;
-                println!("checkpointing..");
+                log::debug!("checkpointing..");
                 self.save()?;
-                println!("new best score: {:?}", self.lowest_error);
+                log::debug!("new best score: {:?}", self.lowest_error);
             }
             // process the checkpoint_window buffer
             self.checkpoint_window.clear();
@@ -569,7 +353,7 @@ impl <'a>NormNet <'a>{
             let error_sum = error.iter().fold(0.0, |acc, x| acc + x)/error.len() as f32;
             self.checkpoint_window.push(error_sum);
 
-            println!("error sum: {}", error_sum);
+            log::debug!("error sum: {}", error_sum);
         }
         // return Ok with Box dyn error
         Ok(())
@@ -580,7 +364,7 @@ impl <'a>NormNet <'a>{
     pub fn save(&mut self) -> Result<(), Box<dyn Error>> {
         // save the model to disk in the current directory
         if self.SavedModelSaver.borrow().is_none() {
-            println!("initializing saved model saver..");
+            log::debug!("initializing saved model saver..");
             let mut all_vars = self.net_vars.clone();
             all_vars.extend_from_slice(&self.minimize_vars);
             let mut builder = tensorflow::SavedModelBuilder::new();
@@ -648,13 +432,13 @@ impl <'a>NormNet <'a>{
         let cur_checkpoint_name = format!("{}/{}_{}", self.name, self.name, uuid);
 
         //update checkpoint_name
-        println!("saving model to {}", cur_checkpoint_name);
+        log::debug!("saving model to {}", cur_checkpoint_name);
         self.SavedModelSaver.borrow_mut().as_mut().unwrap().save(
             &self.session,
             &self.scope.graph(),
             cur_checkpoint_name.clone(),
         )?;
-        println!("serializing non-graph variables to {}", cur_checkpoint_name);
+        log::debug!("serializing non-graph variables to {}", cur_checkpoint_name);
         self.serialize_network(cur_checkpoint_name.clone())?;
         self.checkpoint_name = Some(cur_checkpoint_name);
 
@@ -664,7 +448,7 @@ impl <'a>NormNet <'a>{
     /// load the saved model in the directory dir and restore it in self, removing the 
     /// previous tensorflow graph and session.
     pub fn load(&mut self, dir: String) -> Result<(), Box<dyn Error>> {
-        println!("loading previously saved model..");
+        log::debug!("loading previously saved model..");
         let mut graph = Graph::new();
         //TODO: ensure we can access variables from graph or otherwise
         let bundle =
@@ -713,15 +497,15 @@ impl <'a>NormNet <'a>{
     ) -> Result<Vec<Tensor<f32>>, Box<dyn Error>> {
         //TODO: randomize input and labels while k-folding
         assert_eq!(inputs.len(), labels.len());
-        println!("trainning..");
+        log::debug!("trainning..");
         let mut result = vec![];
 
         let mut input_tensor: Tensor<T> = Tensor::new(&[1u64, inputs[0].len() as u64]);
         let mut label_tensor: Tensor<T> = Tensor::new(&[1u64, labels[0].len() as u64]);
 
-        println!("inputs.len(): {}", inputs.len());
-        println!("{}", inputs[0].len());
-        println!("{}", labels[0].len());
+        log::debug!("inputs.len(): {}", inputs.len());
+        log::debug!("{}", inputs[0].len());
+        log::debug!("{}", labels[0].len());
 
         let mut input_iter = inputs.into_iter();
         let mut label_iter = labels.into_iter();
@@ -772,7 +556,7 @@ impl <'a>NormNet <'a>{
             // update the moving average for time
             let average = avg_t.iter().sum::<f32>() / avg_t.len() as f32;
 
-            println!(
+            log::debug!(
                 "training on {}\n input: {:?} label: {:?} error: {} output: {} seconds/epoch: {:?}",
                 i, input, label, res, output,average 
             );
@@ -803,11 +587,11 @@ impl <'a>NormNet <'a>{
 
         //START OF TRAIN SUBROUTINE
         assert_eq!(inputs.len(), labels.len());
-        println!("trainning..");
+        log::debug!("trainning..");
 
-        println!("inputs.len(): {}", inputs.len());
-        println!("{}", inputs[0].len());
-        println!("{}", labels[0].len());
+        log::debug!("inputs.len(): {}", inputs.len());
+        log::debug!("{}", inputs[0].len());
+        log::debug!("{}", labels[0].len());
 
         let input_itl = inputs.clone();
         let label_itl = labels.clone();
@@ -863,7 +647,7 @@ impl <'a>NormNet <'a>{
             // update the moving average for time
             let average = avg_t.iter().sum::<f32>() / avg_t.len() as f32;
 
-            println!(
+            log::debug!(
                 "training on {}\n input: {:?} label: {:?} error: {} output: {} time/epoch(ms): {:?}",
                 i, input, label, res, output,average 
             );
@@ -874,12 +658,10 @@ impl <'a>NormNet <'a>{
         Ok(())
     }
 
-    //stochastically (with 'selection_pressure=n' for fitness) select checkpointed models 
-    //and save them if they are 'delta_fitness=x' above the previous lowest_error
-    ///stochastically select a model from dir with selection_pressure for the models fitness
+    ///randomly select a model from dir with selection_pressure for the models fitness
     pub fn load_checkpoint_search(&mut self, selection_pressure: f32) -> Result<(), Box<dyn Error>> {
         assert!(selection_pressure >= 0.0 && selection_pressure <= 1.0, "selection_pressure must be between 0 and 1");
-        println!("loading checkpointed models..");
+        log::debug!("loading checkpointed models..");
         let files_iter = fs::read_dir(self.name)?.par_bridge();
         // .par_bridge();
         // create an iterator from files_iter
@@ -943,14 +725,14 @@ impl <'a>NormNet <'a>{
             selection > serialized_network.lowest_error.clone()
         })
         // .inspect(|(dir, serialized_network)| {
-        //     println!("choosing from: {} {:?}", serialized_network.lowest_error, dir);
+        //     log::debug!("choosing from: {} {:?}", serialized_network.lowest_error, dir);
         // })
         .choose(&mut r).context("no checkpoint found")?;
 
         //TODO: search the tree for diversity either by looking at the expected future reward of a node (number of unique paths to unique frontier nodes) 
         //      or searching as horizontally as possible from the current checkpoint node with fitness pressure
 
-        println!("loading network with fitness: {:?}", serialized_network.lowest_error);
+        log::debug!("loading network with fitness: {:?}", serialized_network.lowest_error);
         self.load(dir)?;
         serialized_network.restore(self);
         // TODO: if Root node is selected, reinitialize all variables for random seed of network to ensure robust search
@@ -1023,9 +805,35 @@ impl <'a>NormNet <'a>{
         Ok(outputs.clone())
     }
     //TODO: feedback network and rework name
-    //TODO: feedback network may need some inheritance or builder pattern constructor
+    //TODO: unrolling feedback network
     // pub fn feedback_sequence_evaluate()
+}
 
+///The serialized representation of the model outside of tensorflow graph.
+///Primarily used for checkpoint search meta-learner.
+#[derive(Serialize, Deserialize,Debug)]
+struct SerializedNetwork{
+    /// The previous checkpoint that created this checkpoint for informed graph 
+    /// selection checkpoint search.
+    parent_search_name: String,
+    checkpoint_name: Option<String>,
+    lowest_error: f32,
+} 
+impl SerializedNetwork{
+    /// Constructs a new serialized network from a checkpoint name and the lowest error
+    fn new(parent_search_name: String, checkpoint_name: Option<String>, lowest_error: f32) -> Self{
+        Self{
+            parent_search_name,
+            checkpoint_name,
+            lowest_error,
+        }
+    }
+    fn restore(self, norm_net: &mut Brain) {
+        // let mut Brain = Brain.clone();
+        norm_net.checkpoint_name = self.checkpoint_name;
+        norm_net.lowest_error = self.lowest_error;
+        // Brain
+    }
 }
 
 //TODO: add assertions for more than just runtime error.
@@ -1033,12 +841,12 @@ impl <'a>NormNet <'a>{
 mod tests {
     #[test]
     fn test_net() {
-        println!("test_net");
+        log::debug!("test_net");
         //call the main function
         use crate::*;
 
         //CONSTRUCTION//
-        let mut norm_net = NormNet::new("test_net",2, 1, 10, 10, 10, 1.0, 5 as f32).unwrap();
+        let mut norm_net = Brain::new("test_net",2, 1, 10, 10, norm_layer, 10, 1.0, 5 as f32).unwrap();
 
         //FITNESS FUNCTION//
         //TODO: auto gen labels from outputs and fitness function.
@@ -1063,12 +871,12 @@ mod tests {
     #[test]
     fn test_serialization() {
         //TODO: failing after checkpoint features
-        println!("test_serialization");
+        log::debug!("test_serialization");
         //call the main function
         use crate::*;
 
         //CONSTRUCTION//
-        let mut norm_net = NormNet::new("test_serialization",2, 1, 20, 15, 10, 0.01, 5 as f32).unwrap();
+        let mut norm_net = Brain::new("test_serialization",2, 1, 20, 15, norm_layer, 10, 0.01, 5 as f32).unwrap();
         //TRAIN//
         let mut rrng = rand::thread_rng();
         let mut inputs = Vec::new();
@@ -1102,10 +910,10 @@ mod tests {
             } else {
                 path = is_dir.clone().to_str().unwrap().to_string();
             }
-            println!("{:?}", path);
+            log::debug!("{:?}", path);
         }
         let path = path.to_string();
-        println!("{:?}", path);
+        log::debug!("{:?}", path);
 
         norm_net.load(path.to_string()).unwrap();
 
@@ -1114,10 +922,10 @@ mod tests {
 
     #[test]
     fn test_checkpoint(){
-        println!("test_checkpoint");
+        log::debug!("test_checkpoint");
         use crate::*;
         //CONSTRUCTION//
-        let mut norm_net = NormNet::new("test_checkpoint",2, 1, 200, 96, 10, 10.0, 5 as f32).unwrap();
+        let mut norm_net = Brain::new("test_checkpoint",2, 1, 200, 96, norm_layer, 10, 10.0, 5 as f32).unwrap();
         //TRAIN//
         let mut rrng = rand::thread_rng();
         // create entries for inputs and outputs of xor
@@ -1144,10 +952,10 @@ mod tests {
     }
     #[test]
     fn test_infer(){
-        println!("test_inference");
+        log::debug!("test_inference");
         use crate::*;
         //CONSTRUCTION//
-        let mut norm_net = NormNet::new("test_inference",2, 1, 200, 96, 10, 10.0, 5 as f32).unwrap();
+        let mut norm_net = Brain::new("test_inference",2, 1, 200, 96, norm_layer, 10, 10.0, 5 as f32).unwrap();
         //TRAIN//
         let mut rrng = rand::thread_rng();
         // create entries for inputs and outputs of xor
@@ -1158,10 +966,10 @@ mod tests {
     }
     #[test]
     fn test_evaluate(){
-        println!("test_evaluate");
+        log::debug!("test_evaluate");
         use crate::*;
         //CONSTRUCTION//
-        let mut norm_net = NormNet::new("test_evaluate",2, 1, 200, 96, 10, 1.0, 10 as f32).unwrap();
+        let mut norm_net = Brain::new("test_evaluate",2, 1, 200, 96, norm_layer, 10, 1.0, 10 as f32).unwrap();
         //TRAIN//
         let mut rrng = rand::thread_rng();
 
@@ -1188,4 +996,7 @@ mod tests {
             let res: Tensor<f32> =  norm_net.evaluate(inputs, 1, fitness_function.clone()).unwrap();
         }
     }
+    //TODO: test layer modularity
+    // #[test]
+    // fn test_architectures(){}
 }
